/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ibdutil;

import beagleutil.Samples;
import blbutil.Const;
import blbutil.FileIt;
import blbutil.FileUtil;
import blbutil.InputIt;
import blbutil.IntPair;
import blbutil.SampleFileIt;
import blbutil.Utilities;
import java.io.File;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import vcf.GeneticMap;
import vcf.PlinkGenMap;
import vcf.Marker;
import vcf.RefIt;
import vcf.GTRec;
import vcf.RefGTRec;

/**
 * <p>Class {@code MergeIbdSegments} adds IBD tracts that identifies gaps between
 * two IBD tracts for a pair of individuals. See the {@code usage()}
 * method for usage instructions.
 * </p>
 * <p>See {@code this.usage()} for usage instructions.
 * </p>
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class MergeIbdSegments {

    private static final String program = "merge-ibd-segments.__REV__.jar";

    private static final DecimalFormat DF2 = new DecimalFormat("0.##");
    private static final DecimalFormat DF3 = new DecimalFormat("0.###");

    /**
     * <p>See {@code this.usage()} for usage instructions.
     * </p>
     * @param args command line arguments
     */
    public static void main(String[] args) {
        if (args.length != 4) {
            System.out.println(usage());
            System.exit(0);
        }
        File vcfFile = new File(args[0]);
        File mapFile = new File(args[1]);
        double maxGap = parseAndCheckMaxGap(args[2]);
        int maxDiscord = parseAndCheckMaxDiscord(args[3]);
        Data data = new Data(vcfFile, mapFile);

        try (PrintWriter out = FileUtil.stdOutPrintWriter()) {
            Map<IntPair, List<StringIdSegment>> map = readSegments(data.samples());
            List<StringIdSegment> toMerge = new ArrayList<>();
            for (IntPair key : map.keySet()) {
                List<StringIdSegment> list = map.get(key);
                assert list.isEmpty()==false && toMerge.isEmpty();
                Collections.sort(list);
                int maxEnd = Integer.MIN_VALUE;
                for (int j=0, n=list.size(); j<n; ++j) {
                    StringIdSegment next = list.get(j);
                    if (mergeWithList(toMerge, next, maxEnd, data, maxGap,
                            maxDiscord)==false) {
                        printAndClearList(data, toMerge, out);
                        maxEnd = Integer.MIN_VALUE;
                    }
                    if (next.end() > maxEnd) {
                        maxEnd = next.end();
                    }
                    toMerge.add(next);
                }
                printAndClearList(data, toMerge, out);
            }
        }
    }

    private static double parseAndCheckMaxGap(String gapArg) {
        double maxGap = Double.NaN;
        try {
            maxGap = Double.parseDouble(gapArg);
            if (Double.isFinite(maxGap)==false || maxGap <= 0.0) {
                exitWithError("ERROR: [gap]=" + gapArg);
            }
        }
        catch(NumberFormatException e) {
            exitWithError("ERROR: [gap] is not a numerical value: " + gapArg);
        }
        return maxGap;
    }

    private static int parseAndCheckMaxDiscord(String discordArg) {
        int maxDiscord = -1;
        try {
            maxDiscord = Integer.parseInt(discordArg);
            if (maxDiscord < 0) {
                exitWithError("ERROR: [discord] < 0: " + discordArg);
            }
        }
        catch(NumberFormatException e) {
            exitWithError("ERROR: [discord] is not a nonnegative integer: "
                    + discordArg);
        }
        return maxDiscord;
    }

    private static Map<IntPair, List<StringIdSegment>> readSegments(
            Samples samples) {
        Map<IntPair, List<StringIdSegment>> map = new HashMap<>();
        try (FileIt<String> it = InputIt.fromStdIn()) {
            while (it.hasNext()) {
                String line = it.next().trim();
                if (line.length() > 0) {
                    StringIdSegment segment = StringIdSegment.parse(line);
                    int s1 = segment.sampleIndex1();
                    int s2 = segment.sampleIndex2();
                    if (samples.index(s1) > samples.index(s2)) {
                        String s = "Inconsistency in sample order:" + line;
                        throw new IllegalArgumentException(s);
                    }
                    IntPair key = new IntPair(s1, s2);
                    List<StringIdSegment> list = map.get(key);
                    if (list==null) {
                        list = new ArrayList<>(4);
                        map.put(key, list);
                    }
                    list.add(segment);
                }
            }
        }
        return map;
    }

    private static boolean mergeWithList(List<StringIdSegment> toMerge,
            StringIdSegment next, int maxEnd, Data data, double maxGap,
            int maxDiscord) {
        if (toMerge.isEmpty()) {
            return true;
        }
        else {
            StringIdSegment last = toMerge.get(toMerge.size()-1);
            if (last.chromIndex() != next.chromIndex()) {
                return false;
            }
            else if (next.start() <= maxEnd) {
                return true;
            }
            else {
                int gapStart = data.markerIndex(last.chromIndex(), maxEnd);
                int gapEnd = data.markerIndex(next.chromIndex(), next.start());
                if (gapStart == -1) {
                    exitWithError("ERROR: position missing from VCF file: "
                            + last.chromIndex() + ":" + maxEnd);
                }
                if (gapEnd == -1) {
                    exitWithError("ERROR: position missing from VCF file: "
                            + next.chromIndex() + ":" + next.end());
                }
                assert gapStart <= gapEnd;

                double gapLength = data.genDistance(gapStart, gapEnd);
                if (gapLength > maxGap) {
                    return false;
                }
                else {
                    int s1 = last.sampleIndex1();
                    int s2 = last.sampleIndex2();
                    int discordCnt = data.ibdDiscordCnt(s1, s2, gapStart, gapEnd);
                    return  discordCnt <= maxDiscord;
                }
            }
        }
    }

    private static void printAndClearList(Data data, List<StringIdSegment>
            toMerge, PrintWriter out) {
        assert toMerge.isEmpty()==false;
        boolean isMergedSegment = toMerge.size()>1;
        StringIdSegment first = toMerge.get(0);
        double score = -Double.MAX_VALUE;
        int start = Integer.MAX_VALUE;
        int end = Integer.MIN_VALUE;
        for (int j=0, n=toMerge.size(); j<n; ++j) {
            StringIdSegment seg = toMerge.get(j);
            if (seg.start() < start) {
                start = seg.start();
            }
            if (seg.end()>end) {
                end = seg.end();
            }
            if (seg.score()>score) {
                score = seg.score();
            }
        }
        double startPos = data.genMap().genPos(first.chromIndex(), start);
        double endPos = data.genMap().genPos(first.chromIndex(), end);
        double genLength = endPos - startPos;
        out.print(first.sample1());
        out.print(Const.tab);
        out.print(isMergedSegment ? 0 : first.hap1());
        out.print(Const.tab);
        out.print(first.sample2());
        out.print(Const.tab);
        out.print(isMergedSegment ? 0 : first.hap2());
        out.print(Const.tab);
        out.print(first.chrom());
        out.print(Const.tab);
        out.print(start);
        out.print(Const.tab);
        out.print(end);
        out.print(Const.tab);
        out.print(DF2.format(score));
        out.print(Const.tab);
        out.println(DF3.format(genLength));
        toMerge.clear();
    }

    private static void exitWithError(String msg) {
        System.out.println(usage());
        Utilities.exit(msg);
    }

    private static String usage() {
        StringBuilder sb = new StringBuilder(400);
        sb.append(Const.nl);
        sb.append("usage: cat [in] | java -jar ");
        sb.append(program);
        sb.append(" [vcf] [map] [gap] [discord] > [out]");
        sb.append(Const.nl);
        sb.append(Const.nl);
        sb.append("where");
        sb.append(Const.nl);
        sb.append("  [in]      = IBD output file from Refined IBD analysis");
        sb.append(Const.nl);
        sb.append("  [vcf]     = Phased input VCF file from Refined IBD analysis");
        sb.append(Const.nl);
        sb.append("  [map]     = PLINK format genetic map file with centimorgan (cM) distances");
        sb.append(Const.nl);
        sb.append("  [gap]     = max length of gap between IBD segments (cM)");
        sb.append(Const.nl);
        sb.append("  [discord] = max number of genotypes in IBD gap that are inconsistent with IBD");
        sb.append(Const.nl);
        sb.append("  [out]     = IBD output file with IBD segments after merging");
        sb.append(Const.nl);
        sb.append(Const.nl);
        sb.append("IBD segments for a pair of sample are merged if they overlap or if they are");
        sb.append(Const.nl);
        sb.append("separated by a gap having length <= [gap] cM and having <= [discord] genotypes");
        sb.append(Const.nl);
        sb.append("inconsistent with IBD.  Merged segments will have haplotype indices set to 0");
        sb.append(Const.nl);
        sb.append("and have IBD score equal to the maximal score of the merged segments.");
        sb.append(Const.nl);
        return sb.toString();
    }

    private static final class Data {

        private final Samples samples;
        private final List<GTRec> recs;
        private final Map<IntPair, Integer> indexMap;
        private final GeneticMap genMap;

        /**
         * Constructs a new {@code Data} instance for the specified data.
         * @param vcfFile a VCF file
         * @param mapFile a PLINK format genetic map file with cM distances
         * @throws Exception if a format error is detected in the specified
         * VCF file or genetic map file
         */
        private Data(File vcfFile, File mapFile) {
            this.genMap = PlinkGenMap.fromPlinkMapFile(mapFile);
            this.recs = new ArrayList<>();
            this.indexMap = new HashMap<>(1000000);
            try (FileIt<String> it = InputIt.fromGzipFile(vcfFile);
                    SampleFileIt<RefGTRec> vcfIt = RefIt.create(it)) {
                this.samples = vcfIt.samples();
                int markerCnt = 0;
                while (vcfIt.hasNext()) {
                    RefGTRec rec = vcfIt.next();
                    Marker m = rec.marker();
                    IntPair key = new IntPair(m.chromIndex(), m.pos());
                    indexMap.put(key, markerCnt++);
                    recs.add(rec);
                }
            }
        }

        /**
         * Returns the genetic map.
         * @return the genetic map
         */
        private GeneticMap genMap() {
            return genMap;
        }

        /**
         * Returns the samples
         * @return the samples
         */
        private Samples samples() {
            return samples;
        }

        /**
         * Returns the number of marker.
         * @return the number of markers
         */
        public int nMarkers() {
            return recs.size();
        }

        /**
         * Returns a marker index corresponding to the specified chromosome
         * and position, or {@code -1} if there is no corresponding marker
         * index.
         * @param chrom a chromosome index
         * @param pos a chromosome position
         * @return the marker index corresponding to the specified chromosome
         * and position
         */
        public int markerIndex(int chrom, int pos) {
            IntPair key = new IntPair(chrom, pos);
            Integer value = indexMap.get(key);
            return (value==null) ? -1 : value;
        }

        /**
         * Returns the specified marker.
         * @param marker a marker index
         * @return the specified marker
         * @throws IndexOutOfBoundsException if
         * {@code marker < 0 || marker >= this.nMarkers()}
         */
        private Marker marker(int marker) {
            return recs.get(marker).marker();
        }

        /**
         * Returns the genetic distance between the specified markers.
         * @param startIndex the starting marker index (inclusive)
         * @param endIndex the ending marker index (inclusive)
         * @return the genetic distance between the specified markers
         * @throws IllegalArgumentException if
         * {@code startIndex > endIndex}
         */
        private double genDistance(int startIndex, int endIndex) {
            if (startIndex > endIndex) {
                String s = "startIndex=" + startIndex + " endIndex=" + endIndex;
                throw new IllegalArgumentException(s);
            }
            Marker startMarker = marker(startIndex);
            Marker endMarker = marker(endIndex);
            if (startMarker.chromIndex() != endMarker.chromIndex()) {
                String s = "inconsistent chromosomes: startIndex=" + startIndex
                        + " endIndex=" + endIndex;
                throw new IllegalArgumentException(s);
            }
            return (genMap.genPos(endMarker) - genMap.genPos(startMarker));
        }

        /**
         * Returns the number of genotypes that are inconsistent with
         * IBD for the specified samples between the starting and ending marker
         * indices inclusive.
         * @param sample1 the first sample index
         * @param sample2 the second sample index
         * @param startIndex the starting marker index (inclusive)
         * @param endIndex the ending marker index (inclusive)
         * @return the number of genotypes that are inconsistent with
         * IBD for the specified samples between the specified markers
         * @throws IllegalArgumentException if
         * {@code sample1 < 0 || sample1 >= this.nSamples()}
         * @throws IllegalArgumentException if
         * {@code sample2 < 0 || sample2 >= this.nSamples()}
         * @throws IllegalArgumentException if {@code startIndex > endIndex}
         */
        private int ibdDiscordCnt(int sample1, int sample2, int startIndex,
                int endIndex) {
            if (startIndex > endIndex) {
                String s = "startIndex=" + startIndex + " endIndex=" + endIndex;
                throw new IllegalArgumentException(s);
            }
            int discordCnt = 0;
            for (int m=startIndex; m<=endIndex; ++m) {
                GTRec data = recs.get(m);
                int a1 = data.allele1(sample1);
                int a2 = data.allele2(sample1);
                int b1 = data.allele1(sample2);
                int b2 = data.allele2(sample2);
                if ( (a1!=b1) && (a1!=b2) && (a2!=b1) && (a2!=b2)
                        && (a1!=-1) && (a2!=-1) && (b1!=-1) && (b2!=-1) ) {
                    ++discordCnt;
                }
            }
            return discordCnt;
        }
    }
}
