/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package refined;

import beagleutil.ChromInterval;
import blbutil.Const;
import blbutil.Validate;
import java.io.File;
import java.util.Map;

/**
 * <p>Class {@code RefinedPar} represents the analysis parameters for a
 * refined-ibd analysis.
 * </p>
 * <p>Instances of class {@code RefinedPar} are immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class RefinedPar {

    private final String[] args;

    // data input/output parameters
    private final File gt;
    private final String out;
    private final File excludesamples;
    private final File excludemarkers;
    private final File map;
    private final ChromInterval chromInt;

    // algorithm parameters
    private final int nthreads;
    private final float window;

    // ibd parameters
    private final float lod;
    private final float length;
    private final float scale;
    private final float trim;

    /**
     * Constructs a new {@code RefinedPar} instance from the specified
     * command line arguments.
     * @param args the command line arguments
     * @throws IllegalArgumentException if a command line argument
     * is incorrectly specified
     * @throws NumberFormatException if a numeric value for a parameter
     * is incorrectly specified
     * @throws NullPointerException if {@code args ==  null}
     */
    public RefinedPar(String[] args) {
        int IMAX = Integer.MAX_VALUE;
        float FMIN = Float.MIN_VALUE;
        float FMAX = Float.MAX_VALUE;

        this.args = args.clone();
        Map<String, String> argsMap = Validate.argsToMap(args, '=');

        // data input/output parameters
        gt = Validate.getFile(
                Validate.stringArg("gt", argsMap, true, null, null));
        out = Validate.stringArg("out", argsMap, true, null, null);
        map = Validate.getFile(Validate.stringArg("map", argsMap, false, null, null));
        chromInt = parseChromInt(Validate.stringArg("chrom", argsMap, false, null, null));
        excludesamples = Validate.getFile(
                Validate.stringArg("excludesamples", argsMap, false, null, null));
        excludemarkers = Validate.getFile(
                Validate.stringArg("excludemarkers", argsMap, false, null, null));

        // analysis parameters
        nthreads = modNthreads(Validate.intArg("nthreads", argsMap, false, IMAX, 1, IMAX));
        window = Validate.floatArg("window", argsMap, false, 40f, FMIN, IMAX);
        lod = Validate.floatArg("lod", argsMap, false, 3.0f, FMIN, FMAX);
        length = Validate.floatArg("length", argsMap, false, 1.5f, FMIN, FMAX);
        trim = Validate.floatArg("trim", argsMap, false, 0.15f, 0f, FMAX);
        scale = Validate.floatArg("scale", argsMap, false, 0.0f, 0f, FMAX);

        Validate.confirmEmptyMap(argsMap);
    }

    /**
     * Returns the command line arguments.
     * @return the command line arguments
     */
    public String[] args() {
        return args.clone();
    }

    /**
     * Returns a description of the refined-ibd command line arguments.
     * @return a description of the refined-ibd command line arguments
     */
    public static String usage() {
        String nl = Const.nl;
        return  "Usage: " + RefinedMain.COMMAND + " [arguments]" + nl
                + nl
                + "data input/output parameters ..." + nl
                + "  gt=<VCF file with GT fields>                       (required)" + nl
                + "  out=<output file prefix>                           (required)" + nl
                + "  map=<PLINK map file with cM units>                 (optional)" + nl
                + "  chrom=<[chrom] or [chrom]:[start]-[end]>           (optional)" + nl + nl
                + "  excludesamples=<file with 1 sample ID per line>    (optional)" + nl
                + "  excludemarkers=<file with 1 marker ID per line>    (optional)" + nl + nl

                + "analysis parameters ..." + nl
                + "  nthreads=<number of threads>                       (default: machine-dependent)" + nl
                + "  window=<window cM length>                          (default=40.0)" + nl
                + "  lod=<min LOD score of reported IBD segments>       (default=3.0)" + nl
                + "  length=<min cM length of reported IBD segments>    (default=1.5)" + nl
                + "  trim=<cM end trim when calculating LOD score>      (default=0.15)" + nl
                + "  scale=<model scale factor>                         (default: data-dependent)" + nl;
    }

    private static ChromInterval parseChromInt(String str) {
        ChromInterval ci = ChromInterval.parse(str);
        if (str!=null && str.length()>0 && ci==null) {
            throw new IllegalArgumentException("Invalid chrom parameter: " + str);
        }
        return ci;
    }

    /**
     * Returns the adjusted IBD scale parameter which equal to
     * {@code Math.max(2.0f, (float) Math.sqrt(nSamples/100.0))} if
     * {@code this.ibdscale() == 0f}, and returns
     * {@code this.ibdscale()} otherwise.
     *
     * @param nSamples the number of samples
     * @return a sample-size-adjusted IBD scale parameter if
     * {@code this.ibdscale() == 0f}, and {@code this.ibdscale()} otherwise
     * @throws IllegalArgumentException if {@code nSamples < 0}
     */
    public float adjustedIbdScale(int nSamples) {
        if (nSamples <= 0) {
            throw new IllegalArgumentException(String.valueOf(nSamples));
        }
        if (scale==0) {
            return Math.max(2.0f, (float) Math.sqrt(nSamples/100.0));
        }
        else {
            return scale;
        }
    }

    /**
     * Returns the nthreads parameter, which is equal to
     * {@code Runtime.getRuntime().availableProcessors()} if
     * {@code nthreads == Integer.MAX_VALUE}.
     * @return the nthreads parameter
     */
    private static int modNthreads(int nthreads) {
        if (nthreads==Integer.MAX_VALUE) {
            return Runtime.getRuntime().availableProcessors();
        }
        else {
            return nthreads;
        }
    }

    // data input/output parameters

    /**
     * Returns the gt parameter.
     * @return the gt parameter
     */
    public File gt() {
        return gt;
    }

    /**
     * Returns the out parameter.
     * @return the out parameter
     */
    public String out() {
        return out;
    }

    /**
     * Returns the map parameter.
     * @return the map parameter
     */
    public File map() {
        return map;
    }

    /**
     * Returns the chromosome interval or {@code null} if no chrom
     * parameter was specified.
     *
     * @return the chromosome interval or {@code null} if no chrom
     * parameter was specified.
     */
    public ChromInterval chromInt() {
        return chromInt;
    }

    /**
     * Returns the excludesamples parameter or {@code null}
     * if no excludesamples parameter was specified.
     *
     * @return the excludesamples parameter or {@code null}
     * if no excludesamples parameter was specified
     */
    public File excludesamples() {
        return excludesamples;
    }

    /**
     * Returns the excludemarkers parameter or {@code null}
     * if no excludemarkers parameter was specified.
     *
     * @return the excludemarkers parameter or {@code null}
     * if no excludemarkers parameter was specified
     */
    public File excludemarkers() {
        return excludemarkers;
    }

    // analysis parameters

    /**
     * Returns the nthreads parameter.
     * @return the nthreads parameter
     */
    public int nthreads() {
            return nthreads;
        }

    /**
     * Returns the window parameter.
     * @return the window parameter
     */
    public float window() {
        return window;
    }

    /**
     * Returns the lod parameter.
     * @return the lod parameter
     */
    public float lod() {
        return lod;
    }

    /**
     * Returns the length parameter.
     * @return the length parameter
     */
    public float length() {
        return length;
    }

    /**
     * Returns the trim parameter.
     * @return the trim parameter
     */
    public float trim() {
        return trim;
    }

    /**
     * Returns the scale parameter.
     * @return the scale parameter
     */
    public float scale() {
        return scale;
    }
}
