/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package refined;

import vcf.Markers;
import vcf.Marker;
import blbutil.Const;
import blbutil.FileUtil;
import blbutil.Utilities;
import dag.Dag;
import dag.DagUtil;
import java.io.Closeable;
import java.io.File;
import java.io.PrintWriter;

/**
 * <p>Class {@code RefinedScreenOutput} contains methods for printing
 * screen output for a refined-ibd analysis.
 * </p>
 * <p>Instances of class {@code RefinedScreenOutput} are not thread-safe.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class RefinedScreenOutput implements Closeable {

    private final RefinedPar par;
    private final PrintWriter log;
    private final long startNanos;

    private long markerCnt = 0;
    private long cumDagNanos = 0;
    private long cumIbdNanos = 0;

    /**
     * Constructs a new {@code RunStats} instance and prints
     * analysis start information
     * @param par the analysis parameters
     * @throws NullPointerException if {@code par == null}
     */
    public RefinedScreenOutput(RefinedPar par) {
        this.startNanos = System.nanoTime();
        this.par = par;
        this.log = FileUtil.printWriter(new File(par.out() + ".log"));
        printStartInfo();
    }

    /**
     * Prints initial information about the analysis to a log
     * file and to standard output.
     */
    private void printStartInfo() {
        Utilities.duoPrint(log, RefinedMain.SHORT_HELP + Const.nl);
        Utilities.duoPrintln(log, "Start time: " + Utilities.timeStamp());
        Utilities.duoPrint(log, Utilities.commandLine(RefinedMain.PROGRAM, par.args()));
        if (par.map() == null) {
            String s = Const.nl + "No genetic map is specified: using 1 cM = 1 Mb";
            Utilities.duoPrintln(log, s);
        }
        log.flush();
    }

   /**
     * Prints the number of samples and the model scale
     * @param nSamples the number of samples
     */
    public void printSampleSummary(int nSamples) {
        float ibdScale = par.adjustedIbdScale(nSamples);
        Utilities.duoPrint(log, Const.nl);
        Utilities.duoPrint(log, String.format("Samples:                 %-,11d%n",
                nSamples));
        Utilities.duoPrint(log, String.format("Model scale:             %-11.2f%n",
                ibdScale));
        log.flush();
    }

   /**
     * Prints the window index and chromosome interval
     * @param markers the list of markers in the interval
     * @param windowIndex the window index
     * @throws NullPointerException if {@code markers == null}
     */
    public void printWindow(Markers markers, int windowIndex) {
        Marker first = markers.marker(0);
        Marker last = markers.marker(markers.nMarkers() - 1);
        StringBuilder sb = new StringBuilder(30);
        sb.append(Const.nl);
        sb.append("Window ");
        sb.append(windowIndex);
        sb.append(" (");
        String chr = first.chrom();
        if (chr.equals(Const.MISSING_DATA_STRING)==false) {
            sb.append(chr);
            sb.append(Const.colon);
        }
        sb.append(first.pos());
        sb.append(Const.hyphen);
        if (chr.equals(last.chrom())==false) {
            sb.append(last.chrom());
            sb.append(Const.colon);
        }
        sb.append(last.pos());
        sb.append(')');
        sb.append(Const.nl);
        Utilities.duoPrint(log, sb.toString());
        log.flush();
    }

    /**
     * Adds the specified number of markers to the cumulative marker count.
     * @param nMarkers the number of markers to add
     */
    public void addToMarkerCnt(int nMarkers) {
        markerCnt += nMarkers;
    }

    /**
     * Prints the specified number of markers.
     * @param nMarkers the number of markers
     */
    public void printNMarkers(long nMarkers) {
       Utilities.duoPrint(log, String.format("Markers:                 %-,11d%n",
                nMarkers));
    }

    /**
     * Prints the specified timing results.
     * @param dagNanos the nanoseconds for DAG model construction
     * @param ibdNanos the nanoseconds for IBD segment detection
     * @param totalNanos  the total nanoseconds
     */
    public void printTiming(long dagNanos, long ibdNanos, long totalNanos) {
        cumDagNanos += dagNanos;
        cumIbdNanos += ibdNanos;
        duoPrintNanos("Model construction:      ", dagNanos);
        duoPrintNanos("IBD detection:           ", ibdNanos);
        duoPrintNanos("Total time:              ", totalNanos);
    }

    /**
     * Prints statistics for the specified DAG.
     * @param dag a DAG model
     */
    public void printDagStats(Dag dag) {
        Utilities.duoPrint(log, Const.nl + DagUtil.dagStats(dag));
    }

    /**
     * Closes the log file, releases I/O resources controlled by this object,
     * and prints analysis end information.
     */
    @Override
    public void close() {
        long totalNanos = System.nanoTime() - startNanos;
        Utilities.duoPrint(log, Const.nl);
        Utilities.duoPrintln(log, "Cumulative Statistics:" + Const.nl);
        printNMarkers(markerCnt);
        printTiming(cumDagNanos, cumIbdNanos, totalNanos);
        Utilities.duoPrintln(log, Const.nl + "End time: "
                + Utilities.timeStamp());
        Utilities.duoPrintln(log, RefinedMain.PROGRAM + " finished");
        log.close();
    }

    private void duoPrintNanos(String message, long nanos) {
        Utilities.duoPrint(log, message);
        Utilities.duoPrintln(log, Utilities.elapsedNanos(nanos));
    }
}
