/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package vcf;

import beagleutil.Samples;

/**
 * <p>Class {@code NoPhaseGTWrapper} is a wrapper for a {@code GT}
 * instance that hides all genotype phase data in the wrapped object.
 * </p>
 * <p>Instances of class {@code NoPhaseGTWrapper} are immutable.
 * </p>
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class NoPhaseGTWrapper implements GT {

    private final GT gt;

    /**
     * Constructs a new {@code NoPhaseGTWrapper} instance from the specified
     * data
     * @param gt the genotypes to be wrapped
     * @throws NullPointerException if {@code gt == null}
     */
    public NoPhaseGTWrapper(GT gt) {
        if (gt==null) {
            throw new NullPointerException("gl==null");
        }
        this.gt = gt;
    }

    @Override
    public float gl(int marker, int sample, int a1, int a2) {
        if (a1==a2) {
            return gt.gl(marker, sample, a1, a2);
        }
        else {
            float f1 = gt.gl(marker, sample, a1, a2);
            float f2 = gt.gl(marker, sample, a2, a1);
            return Math.max(f1, f2);
        }
    }

    @Override
    public boolean isPhased() {
        return false;
    }

    @Override
    public boolean isGTData() {
        return gt.isGTData();
    }

    @Override
    public boolean isPhased(int sample) {
        return false;
    }

    @Override
    public boolean isPhased(int marker, int sample) {
        return false;
    }

    @Override
    public int allele1(int marker, int sample) {
        return gt.allele1(marker, sample);
    }

    @Override
    public int allele2(int marker, int sample) {
        return gt.allele2(marker, sample);
    }


    @Override
    public int allele(int marker, int hap) {
        return gt.allele(marker, hap);
    }

    @Override
    public int nMarkers() {
        return gt.nMarkers();
    }

    @Override
    public Marker marker(int marker) {
        return gt.marker(marker);
    }

    @Override
    public Markers markers() {
        return gt.markers();
    }

    @Override
    public int nHaps() {
        return gt.nHaps();
    }

    @Override
    public int nSamples() {
        return gt.nSamples();
    }

    @Override
    public Samples samples() {
        return gt.samples();
    }

    @Override
    public String toString() {
        return gt.toString();
    }
}
